! Copyright 2010-2014 Lixin He, Kun Cao and Zhen Wang
! 
! This file is part of 3tangle
! 
! 3tangle is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
! 
! 3tangle is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
! 
! You should have received a copy of the GNU General Public License
! along with 3tangle.  If not, see <http://www.gnu.org/licenses/>.
! 
!  Subroutine of constrained simulated annealing (CSA)
!
!  Purpose:	Implement constrained simulated annealing (CSA)
!
!  Input:	m = 2*NC^2 + NP + 2, the number of residual functions
!		      n = NP + 2*NP*NC, the number of varibles have to be generated by random
!		      X, the array name of input array X(i)
!		      fvec20, the value of target function f()
!		      IFLAG, the flag of ...
! 
!  Output:	Update the value of fvec20 after one sweep of CSA
!
!=========================================================================================

subroutine CSA(m,n,X,fvec20,IFLAG)
   use minimum
   use para
   use common_doe
   implicit none

!==========Internal parameters=============================
!  t0: initial temperature
!  nsmax: max steps at each temperature
!  alpha: cooling rates

!  ran1, function for generating random number
!  i, sweep counting
   double precision,external::ran1
   integer::i
   
!  real*8,save,allocatable::sigma(:)=1.0d-2
!  integer,save,allocatable::iac(:)=0

!  ns, records how many times CSA has been called
!  FVEC20, the old value of target function
   integer(kind=4),save::ns=0
   double precision::FVEC20

!  u, the random number generated by fuction ran1
!  r0, the weight factor for sigma(i)
!  pi, save the constant PI
!  pu, pv, the upperbound and lowerbound of pac
!  pac, the probability of the new X being accepted
!  beta0, beta1, the constant used for adjusting sigma(i)
!  at, used in Metropolis process
!  DOE, the degree of entanglement, i.e. the E() in the target function
!  R2, the value of the last expression in the target function
!  DFVEC2, no longer used
!  accept, 1 mean the new X is accepted, 0 the opposite
!  dec, no longer used
!  M, N, IFLAG, same meaning as elsewhere
!  X(N), the coefficients in the target function
!  FVEC2, the new value of target function
   real*8 r0, u
   real*8 pu, pv, beta0, beta1, pi,pac
   real*8 at 
   real*8 DOE, R2, DFVEC2
!   real*8 t0,alpha 
   LOGICAL accept,dec
   INTEGER M,N,IFLAG
   double precision X(N), FVEC2

!###############################################################################
!  these two parameters must be defined as common variable and initialize beyond the CSA   
!  double precision,save:: sigma(n)=1.0d-2
!  integer(kind=4),save::iac(n)=0  
   integer::ntrial=100
!###############################################################################
   double precision,allocatable:: X1(:), FVEC(:)

!  EXTERNAL FUN, ENORM2 
   allocate(x1(n), FVEC(m))
!  allocate(sigma(n),iac(n))
   
!  open(1000,file='minimum.dat')
!  open(2000,file='minimumx.dat')
!  write(*,*) "enter CSA"
!  write(*,*)'kappa is', kappa
!  stop
!============the constants=====================================!!should be moved in parameter.f90
   pi=3.141592653589793d0
   pu=0.3d0
   pv=0.2d0
   beta0=2.0d0
   beta1=2.0d0
!==============================================================


!  CALL RANDOM_SEED()

!  generate a trial point (x,lambda)
!  first, we decide to change x, or change lambda
!  CALL RANDOM_NUMBER (r0)  ! generate a random number 0<=r0<1

   ns=ns+1
   do i=1, n

!  the amplitude of changes generate from the Cauchy distribution
      u=ran1(seed_cau)
      r0 = dtan(pi*(u-0.5d0))
      X1(:)=X(:)
      X1(i)=X(i)+r0*sigma(i)
      call NORM(M,N,X1,IFLAG)  !do the normalization for the new X(N)
!  calculate the new Lagrangian
!  write(*,*) "call FUN"
!  call FUN(M,N,X1,FVEC,IFLAG)
!  FVEC2=ENORM2(M,FVEC)
!  if (it.gt.11) IFLAG=1

      call FUN2(M,N,X1,FVEC,FVEC2,DOE,R2,IFLAG)  !calculate the new value of target function

!  Metropolis process starts
!  decide if the new trial point (x1, lambda) is acceptable
      accept=.FALSE.
         if(FVEC2.lt.FVEC20) then
            accept=.TRUE.
         else
            r0=ran1(seed_met) 
!  CALL RANDOM_NUMBER (r0)
            at=DEXP((FVEC20-FVEC2)*beta(mybeta_num))
               if(r0.lt.at) accept=.TRUE.
         end if   

      if(accept) then   !  record the minmum fvec, x ,doe ,r2
         iac(i)=iac(i)+1
         X(:)=X1(:)  ! update X(:)
         FVEC20=FVEC2
      
         if(FVEC20<FVEC2_min)then !if new min of fvec2 appears, update everything
            fvec2_min=fvec20
            x_min(:)=x(:)
            r2_min=r2
            doe_min=doe
         end if
      end if
      tote=fvec20
      
   end do !end fo of i=1,n, one sweep is completed

      if(ns.eq.ntrial) then
         if(adapflag==1) then
            do i=1, n
               pac=real(iac(i))/real(ntrial)  !for every ntrial times, calculate the acceptance rate
     
!  adjust sigma(i) according to the acceptance rate
               if(pac.gt.pu) then
                  sigma(i)=sigma(i)*(1+beta0*(pac-pu)/(1-pu))
               else if (pac.lt.pv) then 
                  sigma(i)=sigma(i)/(1+beta1*(pv-pac)/pv)
               end if
            end do ! i=1,n
         end if !if(adapflag==1)
         ns=0   !reset ns
         iac(:)=0  !reset iac
      end if
        
      continue
     
!  DFVEC2=(FVEC2-FVEC20)
!  write(*,*) "T=", t
!  write(*,*) "it=", it, "DOE=", DOE, "R2=", R2, "ETOT=", FVEC2

!  t=t*alpha   
!  if(R2.lt.1.0d-3) p_lambda=1.0d-2
  
!  deallocate(x1, FVEC)
!  deallocate(sigma,iac)

   return
   end
