! Copyright 2010-2014 Lixin He, Kun Cao and Zhen Wang
! 
! This file is part of 3tangle
! 
! 3tangle is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
! 
! 3tangle is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
! 
! You should have received a copy of the GNU General Public License
! along with 3tangle.  If not, see <http://www.gnu.org/licenses/>.
! 
!  Purpose: Subroutine to calculate the three-way entanglement of a three-particle pure state
!           [Eq. (25) of PRA 61, 052306 (2000).]
!
!  Input:   NLEV, number of single-particle levels
!           NC, number of two-particle configurations
!           cr,ci, coefficents of many-particle configurations
!           entropy, the three-way entanglement of a three-particle pure state
!  Output:  entropy, the three-way entanglement of a three-particle pure state
!  
!========================================================================================

SUBROUTINE EP3(NLEV,NC,cr,ci,entropy)
   integer NLEV, NC
   !double complex  tau      !three-body entanglement 
   double precision cr(NC), ci(NC)
   double complex,allocatable::omega(:,:,:)
   double complex:: d1, d2, d3 !omega
   double precision s(2,2), entropy  !entropy, i.e.,the tau3
   integer p, p1
   data s /0d0, 1.0d0, -1.0d0, 0d0/

   if(NLEV.ne.2) then
      write(*,*) "NLEV /=2, not implemented"
      stop
   end if
   
   if (NC.ne.NLEV*NLEV*NLEV) then
      write(*,*) "NC\= NLEV^3. Something is wrong!"
      stop
   end if
   
   allocate (omega(NLEV,NLEV,NLEV))

   !transfer C to matrix omega 
   id=0
   do ia=1, NLEV
      do ib=1, NLEV
         do ic=1, NLEV
            id=id+1
            omega(ia,ib,ic)=cr(id) + (0.d0,1.0d0)*ci(id)
         end do !ic
      end do !ib
   end do !ia

!      omega(:,:,:)=0.0
!      omega(1,1,1)=0.8
!      omega(2,2,2)=0.6

   d1=omega(1,1,1)*omega(1,1,1)*omega(2,2,2)*omega(2,2,2) &
    + omega(1,1,2)*omega(1,1,2)*omega(2,2,1)*omega(2,2,1) &
    + omega(1,2,1)*omega(1,2,1)*omega(2,1,2)*omega(2,1,2) &
    + omega(2,1,1)*omega(2,1,1)*omega(1,2,2)*omega(1,2,2)

   d2=omega(1,1,1)*omega(2,2,2)*omega(1,2,2)*omega(2,1,1) &
    + omega(1,1,1)*omega(2,2,2)*omega(2,1,2)*omega(1,2,1) &
    + omega(1,1,1)*omega(2,2,2)*omega(2,2,1)*omega(1,1,2) &
    + omega(1,2,2)*omega(2,1,1)*omega(2,1,2)*omega(1,2,1) &
    + omega(1,2,2)*omega(2,1,1)*omega(2,2,1)*omega(1,1,2) &
    + omega(2,1,2)*omega(1,2,1)*omega(2,2,1)*omega(1,1,2)

   d3=omega(1,1,1)*omega(2,2,1)*omega(2,1,2)*omega(1,2,2) & 
    + omega(2,2,2)*omega(1,1,2)*omega(1,2,1)*omega(2,1,1)


!      write(*,*) d1, d2, d3
   entropy= 4.0d0* DSQRT(real((d1-2.0d0*d2+4.0d0*d3) &
       *conjg(d1-2.0d0*d2+4.0d0*d3),8))

!      if(entropy.le.0) write(*,*) "entropy <0"

!      write(*,*) "entropy=", entropy
!      stop
   deallocate (omega)

   RETURN
END 



   
   

    
