! Copyright 2010-2014 Lixin He, Kun Cao and Zhen Wang
! 
! This file is part of 3tangle
! 
! 3tangle is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
! 
! 3tangle is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
! 
! You should have received a copy of the GNU General Public License
! along with 3tangle.  If not, see <http://www.gnu.org/licenses/>.
! 
!  Purpose: Calculate the target function
!
!  Input:   M = 2*NC^2 + NP + 2, the number of residual functions
!		      N = NP + 2*NP*NC, the number of varibles have to be generated by random
!		      X, the array name of input array X(i)
!           DOE, the degree of entanglement, i.e. the E() in the target function
!           R2, the value of the last expression in the target function
!           IFLAG, the flag of ...
!           FVEC2, the value of target function f()
!           FVEC(i), store the elements of DOE and R2
!  Output:  The normalized X(i)
!=================================================================================


SUBROUTINE FUN2(M,N,X,FVEC,FVEC2,DOE,R2,IFLAG)
   use common_doe
   INTEGER M,N,IFLAG
   double precision X(N), FVEC(M), FVEC2, DOE, R2
   double precision,allocatable:: q(:), cr(:,:), ci(:,:),ccr(:), cci(:)   ! q_i^2 = p_i, i=1, NP
   double precision q2tot, crm, dr, di  ! \sum_i q_i^2
               ! \sum_\alpha |C_{i\alpha}|^2
               ! \sum_i q_i^2 c_{i,\alpha}*c^*_{i,\beta}
!  double precision kappa 
   double precision entropy

   allocate(q(NP), cr(NP,NC), ci(NP,NC), ccr(NC), cci(NC))


!  write(*,*) "enter FUN"
!  write(*,*) "number of partilces NPAR=", NPAR
!  write(*,*) "number of single-particle levels NLEV=", NLEV
!  write(*,*) "number of mixed states NP=", NP
 
   do i=1, NP
      q(i)= X(i)
   end do

   do i=1, NP
      do j=1, NC
         i1=(i-1)*NC
         cr(i,j)=X(NP+2*(i1+j)-1) 
         ci(i,j)=X(NP+2*(i1+j))
      end do
   end do

!  now calculate the weight function FVEC()

!  First calculate the entanglement of the states

   DOE=0.d0
   FVEC(:)= 0.d0

   if(doeflag==1)then
      do i=1, NP
         ccr(:)=cr(i,:)
         cci(:)=ci(i,:)
         if(NPAR.eq.2) then
      !  call EP2(NLEV,NC,ccr,cci,entropy) ! calculate the DOE of a two-particle pure state
            if(IFLAG.eq.1) then
               write(*,*) "IFLGA= 1", "i=", i, "entropy=",entropy 
            end if
         else if (NPAR.eq.3) then
            call EP3(NLEV,NC,ccr,cci,entropy) ! calculate the DOE of a three-particle pure state
         else 
            write(*,*) "NPAR/=2,3, not implemented"
         end if
         FVEC(1)= FVEC(1) + q(i)*q(i)*entropy 
      end do
   
      DOE = FVEC(1)
      FVEC(1)=DSQRT(FVEC(1))
   end if  ! the flag whether to inclued DOE in the search process
!    write(*,*) "DOE=", DOE

!     FVEC(1)=0

! sqrt of the mixed state entanglement

! calculate FVEC(2)=\kappa * (\sum_i q_i^2 -1) 

   q2tot=0.0d0
   do i=1, NP
      q2tot = q2tot + q(i)*q(i)
   end do 

!  FVEC(2)=q2tot-1.0d0

!  calculate FVEC(3:NP+2) = \kappa * (\sum_\alpha |C_{i\alpha}|^2-1)

   do i=1, NP
      crm=0.0d0
      do ia=1,NC
         crm=crm+cr(i,ia)*cr(i,ia)+ci(i,ia)*ci(i,ia)
      end do
!    FVEC(i+2)=crm-1.0d0
!    write(*,*) i+2, FVEC(i+2),crm
   end do

!  calculate FVEC(NP+3:NP+2+2*NC^2) = \kappa * ( \sum_i q_i^2* C_{i\alpha} * C^*_{i\beta} -M_{\alpha\beta}

   IC=0
   do ia=1, NC
      do ib=1, NC
         dr=0.0d0
         di=0.0d0
         do i=1, NP
            dr=dr+q(i)*q(i)*(cr(i,ia)*cr(i,ib)+ci(i,ia)*ci(i,ib))
            di=di-q(i)*q(i)*(cr(i,ia)*ci(i,ib)-ci(i,ia)*cr(i,ib))
         end do ! i
         IC=IC+2
         FVEC(NP+1+IC) = (dr - MATR(ia,ib))
         FVEC(NP+2+IC) = (di - MATI(ia,ib))
      end do ! ib
   end do !ia

   FVEC2=0.0d0
   R2=0.0d0
!  rlambda(1)=0.0d0

   FVEC(1)=0.0d0
!  do i=2, NP+2
!     FVEC(1)=FVEC(1)+FVEC(i)*FVEC(i)
!      FVEC(i)=0.0d0
!  end do 

!  do i=1, M-NP-2
   do i=NP+3, M
!     FVEC(i)=DABS(FVEC(i))
!     if(FVEC(i).le.1.0d-3) FVEC(i)=0.0d0
      FVEC2=FVEC2 + kappa*FVEC(i)*FVEC(i)     !FVEC(i)*rlambda(i)+0.5d2*FVEC(i)*FVEC(i)
!     FVEC2=FVEC2+kappa*(1-exp(-FVEC(i)*FVEC(i)/(2.d0*gauss**2)))
      R2=R2+FVEC(i)*FVEC(i)
   end do
   FVEC2=FVEC2+DOE

!  write(*,*) "X=", (X(i),i=1,N) 
!  write(*,*) "FVEC=", (FVEC(i),i=1,M)
!
!   if(IFLAG.eq.1)    
!   write(*,*) "FVEC2=",FVEC2, FVEC(1)
!   write(*,*) "R2=",R2
   deallocate(q, cr, ci,ccr,cci)
   return
   END


