! Copyright 2010-2014 Lixin He, Kun Cao and Zhen Wang
! 
! This file is part of 3tangle
! 
! 3tangle is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
! 
! 3tangle is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
! 
! You should have received a copy of the GNU General Public License
! along with 3tangle.  If not, see <http://www.gnu.org/licenses/>.
! 
!  Purpose: Do the normalization for X(i)
!
!  Input:   M = 2*NC^2 + NP + 2, the number of residual functions
!		      N = NP + 2*NP*NC, the number of varibles have to be generated by random
!		      X, the array name of input array X(i)
!  Output:  The normalized X(i)
!================================================================================

SUBROUTINE NORM(M,N,X,IFLAG)
   use common_doe
!

   INTEGER M,N,IFLAG
   double precision X(N)
   double precision,allocatable::q(:), cr(:,:), ci(:,:), R2(:)   ! q_i^2 = p_i, i=1, NP
   double precision q2tot, crm, dr, di  ! \sum_i q_i^2
               ! \sum_\alpha |C_{i\alpha}|^2
               ! \sum_i q_i^2 c_{i,\alpha}*c^*_{i,\beta}
   double precision entropy

   allocate(q(NP), cr(NP,NC), ci(NP,NC))
   allocate(R2(NP))
 
!  convert X(i) into q(i) and cr(i,j),ci(i,j)
   do i=1, NP
      q(i)= X(i)
   end do

   do i=1, NP
      do j=1, NC
         i1=(i-1)*NC
         cr(i,j)=X(NP+2*(i1+j)-1) 
         ci(i,j)=X(NP+2*(i1+j))
      end do
   end do

!  calculate the sum of q(i)^2
   q2tot=0.0d0
   do i=1, NP
      q2tot = q2tot + q(i)*q(i)
   end do 
   q2tot=DSQRT(q2tot)

!  normalize q(i)
   do i=1, NP
      X(i)=X(i)/q2tot
   end do 
 
!  calculate the sum of cr and ci
   R2(:)=0.0d0
   do i=1, NP
      crm=0.0d0
     do ia=1,NC
        crm=crm+cr(i,ia)*cr(i,ia)+ci(i,ia)*ci(i,ia)
     end do
     R2(i)=DSQRT(crm)
   end do

!  NORMALIZE WAVEFUNCTIONS
   do i=1, NP
      do j=1, NC
         i1=(i-1)*NC
         X(NP+2*(i1+j)-1)=cr(i,j)/R2(i)
         X(NP+2*(i1+j))=ci(i,j)/R2(i)
      end do
   end do

   deallocate(q, cr, ci)
   deallocate(R2)
   return
   END


