! Copyright 2010-2014 Lixin He, Kun Cao and Zhen Wang
! 
! This file is part of 3tangle
! 
! 3tangle is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
! 
! 3tangle is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
! 
! You should have received a copy of the GNU General Public License
! along with 3tangle.  If not, see <http://www.gnu.org/licenses/>.
! 
!  Purpose:   
!  "Minimal" random number generator of Park and Miller with
!  Bays-Durham shuffle and added safeguards. Returns a uniform
!  random deviate between 0.0 and 1.0 (exclusive of the end points).
!  Call with idum a negative integer to initialize; thereafter do
!  not alter idum between successive deviates in a sequence. RNMX
!  should approximate the largest floating point value that is 
!  less than 1.
!
!  Input: idum, the seed
!  Output:  random number between 0.0 and 1.0
!============================================================================
double precision function ran1(idum)
   implicit none
     
   integer::idum     !=-2008
   
   integer ia, im, iq, ir, ntab, ndiv
   real am, eps, rnmx
   
   parameter (ia=16807, im=2147483647, am=1./im)
   parameter (iq=127773, ir=2836, ntab=32, ndiv=1+(im-1)/ntab)
   parameter (eps=1.2e-7, rnmx=1.-eps)

   integer j, k, iv(ntab), iy
   save iv, iy

   data iv /ntab*0/, iy /0/

   if (idum.le.0 .or. iy.eq.0) then
      idum = max(-idum,1)
      do j = ntab+8, 1, -1
         k = idum/iq
         idum = ia*(idum-k*iq)-ir*k
         if (idum .lt. 0) idum = idum + im
         if (j .le. ntab) iv(j) = idum
      enddo
      iy = iv(1)
   endif

   k = idum/iq
   idum = ia*(idum-k*iq) - ir*k
   if (idum.lt.0) idum = idum + im
   j = 1 + iy/ndiv
   iy = iv(j)
   iv(j) = idum
   ran1 = min(am*iy,rnmx)

   return
end


FUNCTION cauchyub()
   use para,only :deltaub,muub
   implicit none
   !INTEGER::idum
   double precision::cauchyub
   double precision,parameter::pi=3.141592653589793d0
   double precision,external::ran1
   integer,save::seedub=-4321

   double precision::scale=1.0
   cauchyub=deltaub*dtan(pi*(ran1(seedub)-0.5))+muub
   cauchyub=cauchyub*scale

end function




!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
