! Copyright 2010-2014 Lixin He, Kun Cao and Zhen Wang
! 
! This file is part of 3tangle
! 
! 3tangle is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
! 
! 3tangle is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
! 
! You should have received a copy of the GNU General Public License
! along with 3tangle.  If not, see <http://www.gnu.org/licenses/>.
! 
module denmatrix
   
   contains
   !=================================================================
   !   
   !  Purpose: Calculate density matrix
   !  Input:   n, the demension of the vector |v>
   !  Output : |v><v|
   !
   !=================================================================
   function denmat(n,v)
      implicit none
      integer::n   !dimension of the vector
   !   complex(kind=8),allocatable,dimension(:,:)::denmat
	  complex(kind=8),dimension(1:n,1:n)::denmat
      complex(kind=8),dimension(1:n)::v
      integer,dimension(1:n)::a
      integer::i,j
      
      if(n/=size(v))then
         write(*,*)'the dimension dismatch'
         stop
      end if
      
   !   allocate(denmat(n,n))
      
      do i=1,n
         do j=1,n
            denmat(i,j)=v(i)*v(j)
         end do
      end do
      
  end function
   
   !===================================================================================
   !  Purpose: Gernerate density matrix of gerneralized GHZ/W mixed state
   !  Input:   p, weight factor of generalized GHZ state
   !           n, demension of the density matrix
   !           dens, density matrix of the mixed state
   !  Output:  Density matrix of the mixed state
   !===================================================================================
   subroutine gen_denmat(p,n,dens)
      implicit none
      integer::n,i,j
      double precision::p,s,p0
      complex(kind=8),dimension(1:n)::v1,v2,v3 ! the vector to construct the density matrix
      complex(kind=8),dimension(1:n,1:n)::dens,e ! the constructed density matrix
      double precision,parameter::a=1/sqrt(2.0), b=1/sqrt(2.0), c=1/sqrt(3.0), d=1/sqrt(3.0), f=1/sqrt(3.0)
!      double precision,parameter::a=0.2,b=sqrt(1.0-a*a),c=0.2,d=0.2,f=sqrt(1.0-c*c-d*d) 
      !
      
      v1=(/(1,0),(0,0),(0,0),(0,0),(0,0),(0,0),(0,0),(1,0)/) ! GHZ state
      v1(1)=(a,0)
      v1(8)=(b,0) !Generalized GHZ state, |gGHZ> = a|000> + b|111>

      v2=(/(0,0),(1,0),(1,0),(0,0),(1,0),(0,0),(0,0),(0,0)/) ! W state
      v2(2)=(c,0)
      v2(3)=(d,0)
      v2(5)=(f,0) !Generalized W state, |gW> = c|001> + d|010> + f|100>
      
      v3=(/(0,0),(0,0),(0,0),(0,0),(0,0),(0,0),(1,0),(0,0)/) !|110>
      
      s=4*c*d*f/a*a*b
      p0=s**(2.d0/3.d0)/(1+s**(2.d0/3.d0))
      !if(my_rank==0)then
      !write(*,*)'s=',s
      !write(*,*)'p0=',p0
      !end if
      !v1=(1.d0/dsqrt(2.d0))*v1
      !v2=(1.d0/dsqrt(3.d0))*v2
      
      dens=p*denmat(n,v1)+(1-p)*denmat(n,v2)
	  !call getden(n, v1, dens)
	  !call getden(n, v2, dentmp)
      !dens=p*dens+(1-p)*dentmp
      
   !   do i=1,n
   !      do j=1,n
   !         if(i==j)then
   !            e(i,j)=(1,0)
   !         else
   !            e(i,j)=(0,0)
   !         end if
   !      end do
   !   end do
      
      !dens=dens+((1.d0-p)/8.d0)*e  ! add white noise using unit matrix
   
   end subroutine
   
   
   !===============================================================================================================
   !  Purpose: Calculate the density matrix by tracing any particle of the 4-particle density matrix, where the 
   !           4-particle state is alpha*|W>+beta|GHZ>
   !  Input:   n, the demension of the density matrix
   !           dens, the density matrix to be calculated
   !           alpha, beta, the coefficients, should satisfy alpha^2+beta^2=1
   !  Output:  dens, the density matrix.
   !===============================================================================================================
   
   subroutine gen_denmat_d(n,alpha,beta,dens)
      implicit none
      integer::n,i,j
      double precision::p
      !complex(kind=8),dimension(1:n)::v1,v2,v3 ! the vector to construct the density matrix
      complex(kind=8),dimension(1:n,1:n)::dens,e ! the constructed density matrix
      complex(kind=8):: alpha,beta 
      
      do i=1,n
         do j=1,n
            if(i==j)then
               e(i,j)=(1,0)
            else
               e(i,j)=(0,0)
            end if
            dens(i,j)=(0.d0,0.d0)
         end do
      end do
      
      !calculate the upper triangle of the density matrix
      dens(1,1)=abs(alpha)**2.d0/4.d0+abs(beta)**2.d0/2.d0
      dens(1,2)=beta*conjg(alpha)/(2.d0*dsqrt(2.d0))
      dens(1,3)=beta*conjg(alpha)/(2.d0*dsqrt(2.d0))
      dens(1,5)=beta*conjg(alpha)/(2.d0*dsqrt(2.d0))
      dens(1,8)=alpha*conjg(beta)/(2.d0*dsqrt(2.d0))
      dens(2,2)=abs(alpha)**2.d0/4.d0
      dens(2,3)=abs(alpha)**2.d0/4.d0
      dens(2,5)=abs(alpha)**2.d0/4.d0
      dens(3,3)=abs(alpha)**2.d0/4.d0
      dens(3,5)=abs(alpha)**2.d0/4.d0
      dens(5,5)=abs(alpha)**2.d0/4.d0
      dens(8,8)=abs(beta)**2.d0/2.d0
      
      !the lower triangle is the conjugate of the upper triangle
      do i=1,n
         do j=1,n
            if(i>j)dens(i,j)=conjg(dens(j,i))
         end do
      end do
      !dens=dens+((1.d0-p)/8.d0)*e
   end subroutine
   
end 
   
   
   
   
