! Copyright 2010-2014 Lixin He, Kun Cao and Zhen Wang
! 
! This file is part of 3tangle
! 
! 3tangle is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
! 
! 3tangle is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
! 
! You should have received a copy of the GNU General Public License
! along with 3tangle.  If not, see <http://www.gnu.org/licenses/>.
! 
!  Purpose: This program is to calculate the entanglement of a mixed state.
!  The entanglement of a mixed state is calculated by find the minium 
!  von Neumann entropy of a over ensembles 
!  $\epsilon=\{ p_i,Y_i \}$ realizing the mixed state.
!  (C. H. Bennett {\it et al.} Phys. Rev. A {\bf 54}, 3824 (1996).)
!
!  In the input file, we need to specify the number of particles NPAR,
!  the number of single particle levels NLEV, number of particitions
!  NP, and the density matirx M(\alpha,\beta).
!
!  For NPAR=2, we use subroutine EP2(...) to calculate the weight function
!
!==============================================================================  

subroutine MIXDOE
use para
use minimum
use common_doe
use denmatrix
implicit none
include 'mpif.h'

!==================================Define variables======================================

integer::i,j,ic
integer::iflag=0,iread,proflag=0
double precision::cnom
!NPAR, number of particles, 
!NLEV, number of single-particle levels
!NP, number of partitions
!NC, number of many-particle configurations

complex(kind=8),allocatable,dimension(:,:)::dens   !generated initial density matrix 
integer NPAR0, NLEV0, NP0
integer N, M   ! number of variables, residus   ! 

!  double precision MATR(1000,1000), MATI(1000,1000) ! real and imag part of density matrix MAT
!  double precision  kappa,gamma   ! weights of residus


double precision mr, mi,qq
double precision delta, FVEC2O, FVEC2, FVEC2N, DFVEC2, DOE, DOE0, R2, tmp
integer NSTEP
LOGICAL restart  

double precision p1
double complex::alpha,beta1
!######################################################
!  added by replica-exchange   
double precision,external::ran1
integer,external::weight
integer istatus(MPI_STATUS_SIZE)
!#######################################################

!  variables needed by MINPACK
!  integer INFO
!  double precision  ENORM2   ! DPMPAR
!  double precision  TOL
double precision,allocatable::X(:), X1(:), FVEC(:)
double precision::fvec20
!  double precision t0, alpha

!  COMMON /STAT/kappa,NPAR, NLEV, NP, NC
!  COMMON /DEN/MATR,MATI

!  record the starting time , the end time and the time used
character(len=8)::data1
character(len=10)::time,zone
integer::num1
integer,dimension(1:8)::values
integer,dimension(1:8)::time1
integer,dimension(1:8)::time2
real::nmin,nsec,nhour

!  COMMON ALLOCATABLE /DEN/ MATR, MATI

double precision::timea,timeb,timec
character(len=3)::filename
integer::filenum1,filenum2,filenum3
filenum1=my_rank/100
filenum2=(my_rank-filenum1*100)/10
filenum3=(my_rank-filenum1*100-filenum2*10)
filename=achar(filenum1+48)//achar(filenum2+48)//achar(filenum3+48)
open(200,file='./time/time'//filename//'.dat')

call date_and_time(data1,time,zone,values)
write(200,*)data1,time,zone
write(200,*)values
write(200,'(10x)')
time1=values   !record the start time

!====================end of definition====================================================

!================input the parameters, read form file "input"=============================

open(35, file="input", form="formatted", status="old")

read(35,*) NPAR, NLEV, NP
read(35,*) kappa, eps0 ,dkappa! , TOL
read(35,*) NSTEP
read(35,*) restart
if(my_rank==0)then !if the process rank is 0, print the input parameters
	write(*,*) "Copyright 2010-2014 Lixin He, Kun Cao and Zhen Wang"
	write(*,*) "3tangle program starts"
  write(*,*) "read in NPAR, NLEV, NP"
  write(*,*) "number of partilces NPAR=", NPAR
  write(*,*) "number of single-particle levels NLEV=", NLEV
  write(*,*) "number of mixed states NP=", NP
end if
100  format(1x, I5)

NC=NLEV**NPAR   ! calculate number of many-particle configurations
N=NP+2*NC*NP   ! calculate total numeber of variables to be optimized
M=2*NC*NC+NP+2  ! the number of residual functions need to be calculated

if(my_rank==0)then
  write(*,*) "number of total variables N=", N
  write(*,*) "number of residual funtions M=", M
end if
!===========================end of input===================================================


!========================Allocate memory for variables=====================================
!  LWA=M*N+5*N+M
!  allocate (MATR(NC,NC), MATI(NC,NC))
!  write(*,*) "allocate memory for density matrix completed" 
!  allocate ( X(N), X1(N), DX0(N), DX1(N), GX0(N), GX1(N), FVEC(M) )

allocate ( X(N), X1(N), FVEC(M) ) 
if(my_rank==0) write(*,*) "allocate memory for X, FVEC, IWA, WA completed"
read(35,*) IREAD
read(35,*) p1
!   read(35,*) dtemp
!   if(my_rank==0) write(*,*)'dtemp=',dtemp
!  dtemp=0.9d0 !calculate the coefficients of the density matrix  !!should be external variables

alpha=dcmplx(dtemp,0.d0)
beta1=dcmplx(dsqrt(1.d0-dtemp**2),0.d0)

MATR(:,:)=0.d0
MATI(:,:)=0.d0

if(IREAD.eq.1) then  !read in the density matrix
  if(my_rank==0) write(*,*) "read in the density matrix"
  do i=1, NC
  !  do j=1, NC
  read(35,*) (MATR(i,j) ,j=1,NC) 
  !  write(*,*) (MATR(i,j), j=1, NC)
  !  end do ! j
  end do  ! i
 
 do i=1, NC
  !  do j=1, NC
  read(35,*) (MATI(i,j) ,j=1,NC)  
  !  write(*,*) (MATI(i,j), j=1, NC)
  !  end do ! j
  end do  ! i

else  !generate the density matrix
  allocate(dens(1:nc,1:nc))
  call gen_denmat(p1,nc,dens)
  !call gen_denmat_d(nc,alpha,beta1,dens)
  do i=1,nc   !store the real and imagine part separately
  do j=1,nc
  matr(i,j)=real(dens(i,j))
  mati(i,j)=aimag(dens(i,j))
  end do
  end do 
end if   !if(IREAD.eq.1)

if(my_rank==0)then   !if the process rank is 0, output the density matrix
  write(*,*)'the real part of the density matrix'
  do i=1,nc
  write(*,'(8f15.6)')(matr(i,j),j=1,nc)
  end do

  write(*,*)'the imaginary part of the density matrix'
  do i=1,nc
  write(*,'(8f15.6)')(mati(i,j),j=1,nc)
  end do
end if  !if(my_rank==0)then
!=========================density matrix initialization complete======================================  

!  if(NPAR.eq.2.and.NLEV.eq.2) then
!  call wootters(MATR,MATI)
!  end if

!  rea in the initial X

!  do i=1, N
!     read(35,*) X(i)  ! For i=1, NP,  X(i) store the q_i
!     i=NP+1, NP+2*NC*NP, X(i) store C(j, \alpha) (including real and imag part) 
!  end do 

if(restart) then   !seems useless
  open(50, file='./configin/cont'//filename//'.dat', form="formatted")
  read(50,*) NPAR0, NLEV0, NP0
  if((NPAR0.ne.NPAR).or.(NLEV0.ne.NLEV).or.(NP0.ne.NP)) then
    write(*,*) "Error: input files does not match!"
    stop 
  end if

  do i=1, N
  read(50,*) X(i)
  end do

else
  if(initialx=='ran')then !generate th initial X(i) from random number
    do i=1,n
    x(i)=ran1(seedx)
    end do
    call NORM(M,N,X,IFLAG)

  else if(initialx=='set')then  !calculate the initial X(i)
    x=0.d0
    x(1)=3.d0*(abs(alpha)**2.0)/4.d0+abs(beta1)**2.d0/2.d0
    x(2)=abs(alpha)**2.d0/4.d0+abs(beta1)**2.d0/2.d0
    x(np+1)=real(beta1/dsqrt(2.d0))/dsqrt(x(1))
    x(np+2)=aimag(beta1/dsqrt(2.d0))/dsqrt(x(1))
    x(np+3)=real(alpha/2.d0)/dsqrt(x(1))
    x(np+4)=aimag(alpha/2.d0)/dsqrt(x(1))
    x(np+5)=real(alpha/2.d0)/dsqrt(x(1))
    x(np+6)=aimag(alpha/2.d0)/dsqrt(x(1))
    x(np+9)=real(alpha/2.d0)/dsqrt(x(1))
    x(np+10)=aimag(alpha/2.d0)/dsqrt(x(1))
    x(np+2*nc+1)=real(alpha/2.d0)/dsqrt(x(2))
    x(np+2*nc+2)=aimag(alpha/2.d0)/dsqrt(x(2))
    x(np+2*nc+15)=real(beta1/dsqrt(2.d0))/dsqrt(x(2))
    x(np+2*nc+16)=aimag(beta1/dsqrt(2.d0))/dsqrt(x(2))
    x(1)=sqrt(x(1))
    x(2)=sqrt(x(2))
  end if !if(initialx=='ran')
end if !if(restart)

call FUN2(M,N,X,FVEC,FVEC2,DOE,R2,IFLAG)  !calculate the target function
fvec20=fvec2
if(my_rank==0)write(*,*)'initial fvec20 is ',fvec20   !if the process rank equals 0, output the target function

if(NPAR.eq.2.OR.NPAR.eq.3) then
  allocate(sigma(n),iac(n))
  allocate(x_min(1:n),fvec2_min1(0:nproc-1))
  sigma=1.0d-2
  iac=0
  fvec2_min=1.d10

  !=================================end of allcolate variables=============================================


  !===================================replica-exchange loop================================================
  open(1200,file='./output/betatun'//filename//'.dat')
  open(1000,file='./output/beta.dat')
  mybeta_num=betanum(my_rank)

  do iequi=1,nequi1 !equilibriation before exchange
  call CSA(m,n,X,fvec20,IFLAG)
  end do


  proflag=0
  ischeme=0
  !   wk=0
  !   wk_temp=0
  do irec=1,nrec
  !  control the exchange times and exchange intervals   !seems useless
  if(irec==1)then
    nexch=nexch1
    nswp=nswp1
  else
    nexch=nexch2
    nswp=nswp2
  end if
  exacpt(:)=1
  excount(:)=1
  !equlibrium sweep

  do iexch=1,nexch

  !  if(iexch==1)timea=MPI_WTIME()  !count the time before equlibriation
  mybeta_num=betanum(my_rank)
  do iswp=1,nswp
  call CSA(m,n,X,fvec20,IFLAG)
  end do

  !  if(iexch==1)timeb=MPI_WTIME()   !count time before exchange 

  if(ischeme>nscheme-1)ischeme=0   !there are only 2 schemes, restart from scheme 0
  ibeta=0
  do while(exbeta(ischeme,ibeta)>=0)
  tarbeta1=exbeta(ischeme,ibeta)
  tarbeta2=tarbeta1+1
  excount(tarbeta1)=excount(tarbeta1)+1    !count the number of exchanges

  !#### local block #####
  if(mybeta_num==tarbeta1)then  !the first one in the pair of exchange
    if(my_rank/=process(tarbeta1))then
      write(*,*)'my_rank is',my_rank,'proess is',process(tarbeta1),'exchange is wrong'
      call MPI_Finalize(ierr)
    end if

    nsend(0)=tote  !send the value of target function
    call MPI_SEND(nsend(0),1,MPI_DOUBLE_PRECISION,0,itag,MPI_COMM_WORLD,ierr)
    !write(*,*)'first send is ok'
  end if


  if(mybeta_num==tarbeta2)then  !the second one in the pair of exchange
    if(my_rank/=process(tarbeta2))then
      write(*,*)'my_rank is',my_rank,'proess is',process(tarbeta2),'exchange is wrong'
      call MPI_Finalize(ierr)
    end if

    nsend(0)=tote  !send the value of target function
    call MPI_SEND(nsend(0),1,MPI_DOUBLE_PRECISION,0,itag,MPI_COMM_WORLD,ierr)
    !write(*,*)'the second send is ok'
  end if

  if(my_rank==0)then   !the first process excutes the exchange
    call MPI_RECV(nrecv(0),1,MPI_DOUBLE_PRECISION,process(tarbeta1),itag,MPI_COMM_WORLD,istatus,ierr)
    call MPI_RECV(nrecv(1),1,MPI_DOUBLE_PRECISION,process(tarbeta2),itag,MPI_COMM_WORLD,istatus,ierr)
    !   write(*,*)'the second recv is ok','nrecv=',nrecv(0),nrecv(1)
    !   write(*,*)beta(tarbeta1),beta(tarbeta2)   
    exflag=weight(nrecv(1),nrecv(0),beta(tarbeta2),beta(tarbeta1))
    !   write(*,*)'exflag is',exflag
    if(exflag==1)then
      betanum_temp=betanum(process(tarbeta1))
      betanum(process(tarbeta1))=betanum(process(tarbeta2))
      betanum(process(tarbeta2))=betanum_temp   !exchange beta
      process_temp=process(tarbeta1)
      process(tarbeta1)=process(tarbeta2)
      process(tarbeta2)=process_temp   !exchange process
      exacpt(tarbeta1)=exacpt(tarbeta1)+1 !count the number of exchange acceptance
    end if
  end if

  call MPI_BARRIER(MPI_COMM_WORLD,ierr)  !Synchronizes processes in a communicator
  !broadcast the vector of betanum, process, and exacpt
  call MPI_BCAST(betanum,nproc,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)   
  call MPI_BCAST(process,nproc,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(exacpt,nproc,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)

  !one pair of exchange is over, switch to the next pair
  ibeta=ibeta+1

  end do   !do while(exbeta(ischeme,ibeta)>=0)
  !one scheme of exchange is over, switch to the next scheme
  ischeme=ischeme+1

  write(1200,*)iexch,betanum(my_rank)

  !   if(iexch==1)timec=MPI_WTIME()   !count time after one exchange

  !   if(iexch==1.and.my_rank==0)then
  !   write(1800,*)'the recursion is',irec
  !   write(1800,*)'the equlibration time is',timeb-timea,', the exchange time is',timec-timeb
  !   end if 
  end do ! end of iexch

  if(my_rank==0)then
    write(*,*)'irec is ',irec,'the acpt distribution is as follows:'
    do i=0,nproc-1
    write(*,*)i,exacpt(i),excount(i)
    end do

    write(*,*)'process is as follows'
    do i=0,nproc-1
    write(*,*)i,process(i)
    end do

    write(*,*)'the used kappa is ',kappa

  end if

  if(irec==srec) proflag=1 
  if(proflag==0)then
    call recursion

    !equlibriation after recursion 
    mybeta_num=betanum(my_rank)
    do iequi=1,nequi2
    call CSA(m,n,X,fvec20,IFLAG)
    end do
  end if

  ! find the minimum R^2 after each one recursion 
  ! to adaptively ajust parameter kappa
  if(kappaflag==0)then
    minlocin(1)=fvec2_min
    minlocin(2)=my_rank
    !  call MPI_Allreduce(minlocin,minlocout,1,MPI_2DOUBLE_PRECISION,MPI_MINLOC,MPI_COMM_WORLD)
    call MPI_REDUCE(minlocin,minlocout,1,MPI_2DOUBLE_PRECISION,MPI_MINLOC,0,MPI_COMM_WORLD)
    call MPI_BCAST(minlocout,2,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,ierr)
    call MPI_BARRIER(MPI_COMM_WORLD,ierr)
    if(my_rank==minlocout(2))then
      write(*,*)'rank_min=',minlocout(2)
      write(*,*)'the min fvec2 until this recursion:',fvec2_min
      write(*,*)'the relevant r2 :',r2_min
      write(*,*)'DOE :',doe_min
    endif

    !judge whether kappa should be adjusted
    !         if(minlocout(2)>eps0)then
    !            kappa=kappa*dkappa
    !         else
    !            kappaflag=1
    !         end if

  end if   !end if of kappaflag  

  end do  !end of irec

  !   output the calculated minimum of DOE at different temperatures
  open(500,file='./minimum/minimum'//filename//'.dat')
  ! if(my_rank==0)then
  !   do i=0,nproc-1
  !  if(my_rank==i)then
  write(500,*)'the minmum value of temperature beta is : beta, 1/beta, fvec2_min, r2_min, doe_min'  
  write(500,*)my_rank,betanum(my_rank),fvec2_min, r2_min ,doe_min
  write(500,*)'the minimum x of temperature beta'
  write(500,*)x_min
  !  end if
  !  call MPI_BARRIER(MPI_COMM_WORLD,ierr)

  !   end do

  !  call MPI_SEND(fvec2_min,1,MPI_DOUBLE_PRECISION,0,itag,MPI_COMM_WORLD,ierr)

  !  if(my_rank==0)then
  !   do i=0,nproc-1
  !  call MPI_RECV(fvec2_min1(i),1,MPI_DOUBLE_PRECISION,i,itag,MPI_COMM_WORLD,istatus,ierr)
  !  if(fvec2_min1(i)<fvec2_min1(0))then
  !  fvec2_min1(0)=fvec2_min1(i)
  !  j=i
  !  end if
  !   end do
  !  end if

  !find the minimum value of fvec2 and its location
  !  if(my_rank==0)write(*,*)'before the final minimum'
  minlocin(1)=fvec2_min
  minlocin(2)=my_rank   
  call MPI_reduce(minlocin,minlocout,1,MPI_2DOUBLE_PRECISION,MPI_MINLOC,0,MPI_COMM_WORLD)
  if(my_rank==0)then 
    write(*,*)'the least rank is',minlocout(2)
    write(500,*)'the least rank is',minlocout(2)
  end if


  !output the last configration for the next simulation
  open(50, file='./configout/cont'//filename//'.dat', form="formatted")
  write(50,*) NPAR, NLEV, NP

  do i=1, N
  write(50,*) X(i)
  end do
  if(my_rank==0)then
    open(3000,file='./output/beta_dis.dat')
    do i=0,nproc-1
    write(3000,*)i,betanum(i),process(i)
    end do
  end if

  !  CALL CSA(m,n,X,IFLAG)
  goto 10

end if !if(NPAR.eq.2.OR.NPAR.eq.3)

10   CONTINUE

close(50)
open(55, file="tmp.dat",status="unknown", form="formatted")   
write(55,*) NPAR, NLEV, NP 

deallocate (X, X1,FVEC )

call date_and_time(data1,time,zone,values)
write(200,*)data1,time,zone
write(200,*)values
time2=values

nmin=(time2(3)-time1(3))*24.0*60.0+(time2(5)-time1(5))*60.0+time2(6)-time1(6)+(time2(7)-time1(7))/60.0
nsec=nmin*60.0
nhour=nmin/60.0
write(200,'(10x)')
write(200,*)'total time used (hour):',nhour
write(200,*)'total time used (min):',nmin
write(200,*)'total time used (sec):',nsec

END !  PROGRAM MIXDOE


!  Calculate the two-norm of a vector
double precision FUNCTION ENORM2(M,FVEC)
integer M  !dimension
double precision FVEC(M), FVEC2

FVEC2=0.0d0
do i=1, M
FVEC2=FVEC2+FVEC(i)*FVEC(i)
end do
ENORM2=FVEC2
RETURN
END

!Judge whether to do the exchange using metropolis method
function weight(tote,tote2,beta,beta2) result(exflag)
implicit none
double precision,external::ran1
integer,save::weight_seed=-5000
double precision,intent(in)::tote,tote2,beta,beta2
integer::exflag
double precision::deltah

deltah=(beta-beta2)*(tote-tote2)
if(deltah>0.0.or.exp(deltah)>ran1(weight_seed))then
  exflag=1
else
  exflag=0
end if
end function













