! Copyright 2010-2014 Lixin He, Kun Cao and Zhen Wang
! 
! This file is part of 3tangle
! 
! 3tangle is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
! 
! 3tangle is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
! 
! You should have received a copy of the GNU General Public License
! along with 3tangle.  If not, see <http://www.gnu.org/licenses/>.
! 
!  Purpose: Adjust beta distribution of the processes
!  
!  The method:
!  beta(i) = beta(i-1) + [(1.0 - weight_factor) + weight_factor * exch_acpt_rate(i-1) * lamda] * (beta(i) - beta(i-1))
!  where lamda = [beta(n-1) - beta(0)]/sum_all_process[exch_acpt_rate(i-1) * (beta(i) - beta(i-1))]
!  where i is the number of the process
!====================================================================

subroutine recursion 
   use para
   implicit none
   include 'mpif.h'
   
   double precision::wk_temp0,sum_temp,lamdam
   double precision,allocatable::beta_temp(:),exacpt_temp(:)
   integer::i,j
   double precision::weigh_scale=0.7d0 !i.e, the weight factor
   
   !open(1000,file='./output/beta.dat')
   allocate(beta_temp(0:nproc-1))
   allocate(exacpt_temp(0:nproc-2))
   
   do i=0,nproc-2 !temporarily save the exchange acceptance
      exacpt_temp(i)=exacpt(i)
   end do
   
   wk_temp0=1.d0*minval(exacpt_temp)	!!find min in an array
   !if(wk_temp0<1.0)wk_temp0=1.0e-5
   wk=wk_temp+wk_temp0
   
   do i=0,nproc-1
      wb(i)=wb_temp(i)+wk_temp0*beta(i)
   !beta(i)=wb(i)/wk
   end do
   wk_temp=wk
   wb_temp(:)=wb(:)
   
   !start beta adjustment
   sum_temp=0.d0
   do i=1,nproc-1
      sum_temp=sum_temp+(beta(i)-beta(i-1))*exacpt(i-1)
   end do
   lamdam=(beta(nproc-1)-beta(0))/sum_temp
   beta_temp(0)=beta(0)
   
   do i=1,nproc-1
      beta_temp(i)=beta_temp(i-1)+(1.0-weigh_scale+weigh_scale*exacpt(i-1)*lamdam)*(beta(i)-beta(i-1))
   end do
   beta(:)=beta_temp(:)
   !end beta adjustment
   
   if(my_rank==1)then   !if the process rank is 1, output the beta value of the recursion
      write(1000,*)'beta value of the',irec,'recursion' 
      do i=0,nproc-1
         write(1000,'(i5,f20.5)')i,beta(i)
      end do
   end if
   
   if(irec==nrec)then   !for the last time, adjust beta in a different way to remove the statistical noise
      beta(:)=wb(:)/wk
      if(my_rank==1)then
         write(1000,*)'beta value of the last ',irec,'recursion' 
         do i=0,nproc-1
           write(1000,'(i5,f20.5)')i,beta(i)
         end do
      end if
   end if
   
   ! close(1000)
   end subroutine
   
   
