! Copyright 2010-2014 Lixin He, Kun Cao and Zhen Wang
! 
! This file is part of 3tangle
! 
! 3tangle is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
! 
! 3tangle is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
! 
! You should have received a copy of the GNU General Public License
! along with 3tangle.  If not, see <http://www.gnu.org/licenses/>.
! 
!  Purpose: Do the initialization for replica-exchange
!
!=============================================================  


subroutine replica_init
   use para
   implicit none
   integer::i,j,k
   integer::inttemp
   
   allocate(beta(0:nproc-1))
   allocate(exacpt(0:nproc-1))
   allocate(excount(0:nproc-1))
   allocate(process(0:nproc-1))
   allocate(betanum(0:nproc-1))
   allocate(wb(0:nproc-1))
   allocate(wb_temp(0:nproc-1))

   !initialize beta for all processes
   if(initial_beta==0)then 
      do i=0,nproc-1
         beta(i)=betamin+i*(betamax-betamin)/(nproc-1) !equal space between neighbor beta
      end do
   else  !if(initial_beta/=0)then read beta values form file
      open(600,file='beta_init.dat')
      do i=0,nproc-1
         read(600,*)k,beta(i)
      end do
      close(600)
   end if
   
   !the initial beta distribution
   if(initial_bedis==0)then   !distribute beta in sequence order
      do j=0,nproc-1
         process(j)=j
         betanum(j)=j
      end do
   else !read the distribution form file
      open(5000,file='beta_dis.dat')
      do i=0,nproc-1
         read(5000,*)inttemp,betanum(i),process(i)
         if(inttemp/=i)then
            write(*,*)'the initial beta and process distribution is error'
            call abort
         end if
      end do
   
      !check whether the distribution is right
      do i=0,nproc-1
         if(betanum(process(i))/=i)then
            write(*,*)'the initial beta and process distribution dismatch'
            call abort
         end if
      end do
   
      close(5000)
      write(*,*)'reading the initial beta and process distribution is ok'
   end if !if(initial_bedis==0)
   
   !the initialization of the recursion 
   wk_temp=0.0
   wb_temp(:)=0.0
   
   call ex_scheme
   
   end subroutine
   
   
!determine the scheme of exchange beta
subroutine ex_scheme
   use para
   implicit none
   integer::i,ibeta0,j


   !processes with exbeta==-1 will not do the exchange
   !e.g, if nscheme is 2, the exbeta distribution is as follows
   !scheme0:   0  2  4  6  ...   -1 -1 -1
   !scheme1:   1  3  5  7  ...   -1 -1 -1
   !exchanges happen between pairs from the two schemes 
   allocate(exbeta(0:nscheme-1,0:nproc-1))
   exbeta(:,:)=-1 
   do i=0,nscheme-1
      ibeta0=i
      j=0
      do while(ibeta0<(nproc-1))
         exbeta(i,j)=ibeta0
         ibeta0=ibeta0+nscheme
         j=j+1
      end do
   end do
   
   !if the process rank is 0(the first process), output the scheme
   if(my_rank==0)then   
      do i=0,nscheme-1
         write(*,*)i,'scheme'
         do j=0,nproc-1
            write(*,*)exbeta(i,j)
         end do
      end do
   end if
   
   end subroutine
   

